/****************************************************************************

 ■作例①：ミニAM/FMラジオ (Si4730 Module)
		Copyright(C) 2019 電子うさぎ https://電子うさぎ.com/archives/14953

****************************************************************************/

// PIC16F1705 Configuration Bit Settings

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection Bits (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = SWDTEN    // Watchdog Timer Enable (WDT controlled by the SWDTEN bit in the WDTCON register)
#pragma config PWRTE = OFF      // Power-up Timer Enable (PWRT disabled)
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config BOREN = ON       // Brown-out Reset Enable (Brown-out Reset enabled)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)
#pragma config IESO = ON        // Internal/External Switchover Mode (Internal/External Switchover Mode is enabled)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is enabled)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config PPS1WAY = ON     // Peripheral Pin Select one-way control (The PPSLOCK bit cannot be cleared once it is set by software)
#pragma config ZCDDIS = ON      // Zero-cross detect disable (Zero-cross detect circuit is disabled at POR)
#pragma config PLLEN = OFF      // Phase Lock Loop enable (4x PLL is enabled when software sets the SPLLEN bit)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config BORV = HI        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), high trip point selected.)
#pragma config LPBOR = ON       // Low-Power Brown Out Reset (Low-Power BOR is enabled)
#pragma config LVP = ON         // Low-Voltage Programming Enable (Low-voltage programming enabled)

#include <xc.h>
#include "typedefs.h"
#include "Si47xx/commanddefs.h"
#include "Si47xx/propertydefs.h"

//=================================================================

#define ADDR_SI4730		0x63

#define PORT_LED		LATC5
#define PORT_SD			LATC4
#define PORT_RST		LATA2

#define KEY_NONE		0x00
#define KEY_PREV		0x01
#define KEY_NEXT		0x02
#define KEY_BAND		0x03
#define KEY_SEEK_PREV	0x11
#define KEY_SEEK_NEXT	0x12
#define KEY_POWER		0x04

#define CHAN_NUM_FM		(sizeof(s_wFreqsFM) / sizeof(WORD))
#define CHAN_NUM_AM		(sizeof(s_wFreqsAM) / sizeof(WORD))
#define MAX_VOLUME		21

//=================================================================

static void InitSystem(void);
static void InitApp(void);
static void Run(void);
static void Normal(void);
static void Standby(void);

static BYTE GetKey(void);
static BYTE GetKeySub(void);
static WORD GetSWV(void);
static void SwitchBand(void);
static void PrevFreq(void);
static void NextFreq(void);
static WORD GetPrevFreq(WORD* pFreq, INT8 nCount, WORD wCur);
static WORD GetNextFreq(WORD* pFreq, INT8 nCount, WORD wCur);
static void Seek(BOOL bUp);

static void InitSi47xx(void);
static void EnableRecv(BOOL bEnable);
static void SetProp(WORD wNum, WORD wVal);
static void SetFreqFM(WORD wFreq);
static void SetFreqAM(WORD wFreq);
static WORD SeekFM(BOOL bUp);
static WORD SeekAM(BOOL bUp);
static void SetVolume(BYTE bVol);
static void Si47xxWaitSTC(void);
static void Si47xxCmd(int nSize, int nRes);

static void WriteI2C(BYTE nAddr, BYTE* pData, BYTE nCount);
static void ReadI2C(BYTE nAddr, BYTE* pData, BYTE nCount);

static void SleepDelay(WORD uWdtPs);

//=================================================================

static WORD s_wFreqsFM[] = { 765, 789, 802, 851, 881, 886, 906, 911, 919, 926, 933, 971, 1065 };
static WORD s_wFreqsAM[] = { 558, 666, 828, 1008, 1143, 1179, 1314 };

static BYTE s_nFunc;
static WORD s_wFreqFM;
static WORD s_wFreqAM;
static BYTE s_nVolume;

static BYTE s_Cmd[8];
static BYTE s_Res[16];

//=================================================================

// ●メインエントリ
void main(void)
{
	InitSystem();
	InitApp();
	Run();
}

// ●システム初期化
static void InitSystem(void)
{
	// クロック
	OSCCONbits.IRCF = 0b0000; // 31 kHz LF

	// ポート初期化
	PORTA = PORTC = 0x00;
	LATA = LATC = 0x00;	
	
	// PORT_LED(RC5)
    ANSELCbits.ANSC5 = 0;
	TRISCbits.TRISC5 = 0;

	// PORT_SD(RC4)
	PORT_SD = 1;
    ANSELCbits.ANSC4 = 0;
	TRISCbits.TRISC4 = 0;

	// PORT_RST(RA2)
	PORT_RST = 1;
    ANSELAbits.ANSA2 = 0;
	TRISAbits.TRISA2 = 0;

	// SW(RC3)
	TRISCbits.TRISC3 = 1;

	// Q3(RA5)
	TRISAbits.TRISA5 = 1;

	// I2C(Master mode)
	RC1PPS = 0b10001; // Rxy source is SDA
	RC0PPS = 0b10000; // Rxy source is SCK/SCL

	ANSELCbits.ANSC1 = 0; // RC1/SDA1(9pin)
	ANSELCbits.ANSC0 = 0; // RC0/SCL1(10pin)

    SSP1ADD = 0x00; // 500÷4=125KHz
	SSP1STAT= 0b00000000;

    SSP1CON1bits.SSPM = 0b1000; // I2C Master mode
    SSP1CON1bits.SSPEN = 1;
	
	// ADC
	ADCON1bits.ADPREF = 0; // VREF+ is connected to VDD
	ADCON1bits.ADNREF = 0; // VREF- is connected to AVSS
	ADCON1bits.ADCS = 0b111; // dedicated RC oscillator
	ADCON1bits.ADFM = 1; // Right justified. 
	ADCON0bits.CHS = 0b00111; // AN7
	ADCON0bits.ADON = 1;

	// 割込み許可
	PEIE = 1;
	GIE = 1;

	SleepDelay(0b00111); // 128ms
}

// ●アプリ初期化
static void InitApp(void)
{
	s_nFunc = POWER_UP_IN_FUNC_FMRX;
	s_wFreqFM = s_wFreqsFM[0];
	s_wFreqAM = s_wFreqsAM[0];
	s_nVolume = MAX_VOLUME;
}

// ●実行ループ
static void Run(void)
{
	for(;;)
	{
		Normal();
		Standby();
	}
}

// ●起動状態
static void Normal(void)
{
	PORT_LED = 1;

	ADCON0bits.ADON = 1;

	// 受信動作開始
	TRISAbits.TRISA5 = 0; // FET(Q3) ON
	SleepDelay(0b00110); // 64ms
	InitSi47xx();
	
	EnableRecv(TRUE);
	PORT_SD = 0;
	
	for(;;)
	{
		BYTE nKey = GetKey();
		switch(nKey)
		{
		default:
			// 操作なし
			break;

		case KEY_BAND:
			SwitchBand();
			break;

		case KEY_PREV:
			PrevFreq();
			break;

		case KEY_NEXT:
			NextFreq();
			break;

		case KEY_SEEK_PREV:
			Seek(FALSE);
			while(GetKey() != KEY_NONE)
				SleepDelay(0b00011); // 8ms
			break;

		case KEY_SEEK_NEXT:
			Seek(TRUE);
			while(GetKey() != KEY_NONE)
				SleepDelay(0b00011); // 8ms
			break;

		case KEY_POWER:
			return;
		}
	}
}

// ●スタンバイ
static void Standby(void)
{
	PORT_LED = 0;

	// 受信動作終了
	PORT_SD = 1;
	EnableRecv(FALSE);
	TRISAbits.TRISA5 = 1; // FET(Q3) OFF
	
	// 電源SW押下待ち
	ADCON0bits.ADON = 0;

	// SW(RC3)→Digital
    ANSELCbits.ANSC3 = 0;
	
	IOCCPbits.IOCCP3 = 1;
	IOCIE = 1;
	
	SLEEP();

	IOCIE = 0;
	IOCCPbits.IOCCP3 = 0;

	// SW(RC3)→Analog
    ANSELCbits.ANSC3 = 1;
}

//-----------------------------------------------------------------

// ●キー入力
static BYTE GetKey(void)
{
	BYTE nKey1 = GetKeySub();
	if (nKey1 == KEY_NONE)
		return KEY_NONE;
	
	int i;
	for(i = 0; i < 35; i++)
	{
		BYTE nKey2 = GetKeySub();
		if (nKey1 == nKey2)
			continue;
		
		return (i > 2)? nKey1 : KEY_NONE;
	}
	
	if (nKey1 == KEY_PREV)
		return KEY_SEEK_PREV;
	if (nKey1 == KEY_NEXT)
		return KEY_SEEK_NEXT;
	
	while(GetKeySub() != KEY_NONE);
	
	return KEY_NONE;
}

// ●キー入力
static BYTE GetKeySub(void)
{
	WORD swv = GetSWV();
	// NONE: 1023
	// PREV:  717 
	// NEXT:  618 
	// BAND:  536 
	// POWER:   1
	
	if (swv >= 900)
		return KEY_NONE;
	if (swv > 667)
		return KEY_PREV;
	if (swv > 577)
		return KEY_NEXT;
	if (swv > 100)
		return KEY_BAND;

	return  KEY_POWER;
}

// ●SW電圧値取得
static WORD GetSWV(void)
{
	ADCON0bits.GO = 1;
	while(ADCON0bits.GO_nDONE);
	
	return ADRES;
}

// ●バンド切り替え
static void SwitchBand(void)
{
	PORT_SD = 1;
	EnableRecv(FALSE);

	if (s_nFunc == POWER_UP_IN_FUNC_FMRX)
		s_nFunc = POWER_UP_IN_FUNC_AMRX;
	else
		s_nFunc = POWER_UP_IN_FUNC_FMRX;
				
	EnableRecv(TRUE);
	PORT_SD = 0;
}

// ●前の周波数へ
static void PrevFreq(void)
{
	PORT_SD = 1;
	
	if (s_nFunc == POWER_UP_IN_FUNC_FMRX)
	{
		s_wFreqFM = GetPrevFreq(s_wFreqsFM, CHAN_NUM_FM, s_wFreqFM);
		SetFreqFM(s_wFreqFM);
	}
	else
	{
		s_wFreqAM = GetPrevFreq(s_wFreqsAM, CHAN_NUM_AM, s_wFreqAM);
		SetFreqAM(s_wFreqAM);
	}
	
	PORT_SD = 0;
}

// ●次の周波数へ
static void NextFreq(void)
{
	PORT_SD = 1;
	
	if (s_nFunc == POWER_UP_IN_FUNC_FMRX)
	{
		s_wFreqFM = GetNextFreq(s_wFreqsFM, CHAN_NUM_FM, s_wFreqFM);
		SetFreqFM(s_wFreqFM);
	}
	else
	{
		s_wFreqAM = GetNextFreq(s_wFreqsAM, CHAN_NUM_AM, s_wFreqAM);
		SetFreqAM(s_wFreqAM);
	}
	
	PORT_SD = 0;
}

// ●前の周波数を取得
static WORD GetPrevFreq(WORD* pFreq, INT8 nCount, WORD wCur)
{
	int i;
	for(i = nCount - 1; i >= 0; i--)
	{
		if (pFreq[i] < wCur)
			return pFreq[i];
	}
	return pFreq[nCount - 1];
}

// ●次の周波数を取得
static WORD GetNextFreq(WORD* pFreq, INT8 nCount, WORD wCur)
{
	int i;
	for(i = 0; i < nCount; i++)
	{
		if (pFreq[i] > wCur)
			return pFreq[i];
	}
	return pFreq[0];
}

// ●シーク
static void Seek(BOOL bUp)
{
	PORT_SD = 1;
	
	if (s_nFunc == POWER_UP_IN_FUNC_FMRX)
		s_wFreqFM = SeekFM(bUp);
	else
		s_wFreqAM = SeekAM(bUp);
	
	PORT_SD = 0;
}

//-----------------------------------------------------------------

// ●Si4730初期化
static void InitSi47xx(void)
{
	PORT_RST = 0;
	SleepDelay(0); // tSRST=100us
	PORT_RST = 1;
	SleepDelay(0); // tSRST=30ns
}

// ●受信開始終了
static void EnableRecv(BOOL bEnable)
{
	if (bEnable)
	{
		// POWER_UP
		s_Cmd[0] = POWER_UP;
		s_Cmd[1] = POWER_UP_IN_XOSCEN | s_nFunc;
		s_Cmd[2] = POWER_UP_IN_OPMODE_RX_ANALOG;
		Si47xxCmd(3, 0);		

		if (s_Res[0] & ERR)
			return;

		// GET_REV
		//s_Cmd[0] = GET_REV;
		//Si47xxCmd(1, 15);

		// SET_PROPERTY
		SetProp(FM_DEEMPHASIS, FM_DEEMPHASIS_50US);
		SetProp(FM_BLEND_STEREO_THRESHOLD, 127); // Force mono
		SetProp(FM_BLEND_MONO_THRESHOLD, 127); // Force mono

		SetProp(AM_SEEK_BAND_BOTTOM, 522); // 522KHz
		SetProp(AM_SEEK_BAND_TOP, 1620); // 1620KHz
		SetProp(AM_SEEK_FREQ_SPACING, 9); // 9KHz
		SetProp(AM_CHANNEL_FILTER, AM_CHANNEL_FILTER_AMPLFLT | AM_CHANNEL_FILTER_BW_4KHZ);
		
		// There is a debug feature that remains active in Si4704/05/3x-D60 firmware which can create periodic noise in 
		// audio. Silicon Labs recommends you disable this feature by sending the following bytes (shown here in 
		// hexadecimal form): 
		// 0x12 0x00 0xFF 0x00 0x00 0x00
		SetProp(0xFF00, 0x0000);
		
		// 周波数設定
		if (s_nFunc == POWER_UP_IN_FUNC_FMRX)
			SetFreqFM(s_wFreqFM);
		else
			SetFreqAM(s_wFreqAM);

		// ボリューム設定
		SetVolume(s_nVolume);
	}
	else
	{
		// POWER_DOWN
		s_Cmd[0] = POWER_DOWN;
		Si47xxCmd(1, 0);		
	}
}

// ●プロパティー設定
static void SetProp(WORD wNum, WORD wVal)
{
	s_Cmd[0] = SET_PROPERTY;
	s_Cmd[1] = 0;
	s_Cmd[2] = (BYTE)(wNum >> 8);
	s_Cmd[3] = (BYTE)(wNum & 0x00FF);
	s_Cmd[4] = (BYTE)(wVal >> 8);
	s_Cmd[5] = (BYTE)(wVal & 0x00FF);
	Si47xxCmd(6, 0);
}

// ●周波数設定(FM)
static void SetFreqFM(WORD wFreq)
{
	wFreq *= 10;
	
	// FM_TUNE_FREQ
	s_Cmd[0] = FM_TUNE_FREQ;
	s_Cmd[1] = 0;
	s_Cmd[2] = (BYTE)(wFreq >> 8);
	s_Cmd[3] = (BYTE)(wFreq & 0x00FF);
	s_Cmd[4] = 0;
	Si47xxCmd(5, 0);
	Si47xxWaitSTC();
	
	// FM_TUNE_STATUS
	//s_Cmd[0] = FM_TUNE_STATUS;
	//s_Cmd[1] = FM_TUNE_STATUS_IN_INTACK;
	//Si47xxCmd(2, 6);	
}

// ●周波数設定(AM)
static void SetFreqAM(WORD wFreq)
{
	// AM_TUNE_FREQ
	s_Cmd[0] = AM_TUNE_FREQ;
	s_Cmd[1] = 0;
	s_Cmd[2] = (BYTE)(wFreq >> 8);
	s_Cmd[3] = (BYTE)(wFreq & 0x00FF);
	s_Cmd[4] = 0;
	Si47xxCmd(5, 0);
	Si47xxWaitSTC();
	
	// AM_TUNE_STATUS
	//s_Cmd[0] = AM_TUNE_STATUS;
	//s_Cmd[1] = AM_TUNE_STATUS_IN_INTACK;
	//Si47xxCmd(2, 6);		
}

// ●シーク(FM)
static WORD SeekFM(BOOL bUp)
{
	// FM_SEEK_START
	s_Cmd[0] = FM_SEEK_START;
	s_Cmd[1] = FM_SEEK_START_IN_WRAP;
	if (bUp) s_Cmd[1] |= FM_SEEK_START_IN_SEEKUP;
	Si47xxCmd(2, 0);
	Si47xxWaitSTC();

	// FM_TUNE_STATUS
	s_Cmd[0] = FM_TUNE_STATUS;
	s_Cmd[1] = FM_TUNE_STATUS_IN_INTACK;
	Si47xxCmd(2, 6);	
	
	WORD wFreq = (s_Res[2] << 8) | s_Res[3];
	return wFreq / 10;
}

// ●シーク(AM)
static WORD SeekAM(BOOL bUp)
{
	// AM_SEEK_START
	s_Cmd[0] = AM_SEEK_START;
	s_Cmd[1] = AM_SEEK_START_IN_WRAP;
	if (bUp) s_Cmd[1] |= AM_SEEK_START_IN_SEEKUP;
	s_Cmd[2] = 0;
	s_Cmd[3] = 0;
	s_Cmd[4] = 0;
	s_Cmd[5] = 0;
	Si47xxCmd(6, 0);
	Si47xxWaitSTC();

	// AM_TUNE_STATUS
	s_Cmd[0] = AM_TUNE_STATUS;
	s_Cmd[1] = AM_TUNE_STATUS_IN_INTACK;
	Si47xxCmd(2, 6);	
	
	WORD wFreq = (s_Res[2] << 8) | s_Res[3];
	return wFreq;
}

// ●ボリューム設定
static void SetVolume(BYTE bVol)
{
	// 0～21(→0～63)
	SetProp(RX_VOLUME, bVol * 3);
}

// ●STCINTを待つ
static void Si47xxWaitSTC(void)
{
	// STC割り込み待ち
	//WaitSTCInt();
	
	s_Cmd[0] = GET_INT_STATUS;
	do {
		SleepDelay(0b00110); // 64ms
		Si47xxCmd(1, 0);
	}
	while ((s_Res[0] & STCINT) == 0);
}

// ●コマンド送信
static void Si47xxCmd(int nSize, int nRes)
{
	WriteI2C(ADDR_SI4730, s_Cmd, nSize);

	// CTS割り込み待ち
	//WaitCTSInt();

	do {
		SleepDelay(0b00001); // 2ms
		ReadI2C(ADDR_SI4730, s_Res, 1 + nRes);
	}
	while((s_Res[0] & CTS) == 0);
}

//-----------------------------------------------------------------

// ●I2C送信
static void WriteI2C(BYTE nAddr, BYTE* pData, BYTE nCount)
{
	// Start conditon
	PIR1bits.SSP1IF = 0;
	SSP1CON2bits.SEN = 1;
	while (!PIR1bits.SSP1IF);

	// 7bit address & W flag
	PIR1bits.SSP1IF = 0;
	SSP1BUF = (nAddr << 1) | 0x00;
	while (!PIR1bits.SSP1IF);

	// data
	for (BYTE i = 0; i < nCount; i++)
	{
		PIR1bits.SSP1IF = 0;
		SSP1BUF = pData[i];
		while (!PIR1bits.SSP1IF);
	}

	// Stop condition
	PIR1bits.SSP1IF = 0;
	SSP1CON2bits.PEN = 1;
	while (!PIR1bits.SSP1IF);
}

// ●I2C受信
static void ReadI2C(BYTE nAddr, BYTE* pData, BYTE nCount)
{
	// Start conditon
	PIR1bits.SSP1IF = 0;
	SSP1CON2bits.SEN = 1;
	while (!PIR1bits.SSP1IF);

	// 7bit address & R flag
	PIR1bits.SSP1IF = 0;
	SSP1BUF = (nAddr << 1) | 0x01;
	while (!PIR1bits.SSP1IF);

	// data
	for (BYTE i = 0; i < nCount; i++)
	{
		PIR1bits.SSP1IF = 0;
		SSP1CON2bits.RCEN = 1;
		while (!PIR1bits.SSP1IF);
		pData[i] = SSP1BUF;

		// ACK or NACK
		PIR1bits.SSP1IF = 0;
		SSP1CON2bits.ACKDT = (i == (nCount - 1));
		SSP1CON2bits.ACKEN = 1;
		while (!PIR1bits.SSP1IF);
	}

	// Stop condition
	PIR1bits.SSP1IF = 0;
	SSP1CON2bits.PEN = 1;
	while (!PIR1bits.SSP1IF);
}

//-----------------------------------------------------------------

// ●スリープして待つ
static void SleepDelay(WORD uWdtPs)
{
	/*
	10010  = 1:8388608 (223) (Interval 256s nominal)
	10001  = 1:4194304 (222) (Interval 128s nominal)
	10000  = 1:2097152 (221) (Interval 64s nominal)
	01111  = 1:1048576 (220) (Interval 32s nominal)
	01110  = 1:524288 (219) (Interval 16s nominal)
	01101  = 1:262144 (218) (Interval 8s nominal)
	01100  = 1:131072 (217) (Interval 4s nominal)
	01011  = 1:65536  (Interval 2s nominal)
	01010  = 1:32768 (Interval 1s nominal)
	01001  = 1:16384 (Interval 512 ms nominal)
	01000  = 1:8192 (Interval 256 ms nominal)
	00111  = 1:4096 (Interval 128 ms nominal)
	00110  = 1:2048 (Interval 64 ms nominal)
	00101  = 1:1024 (Interval 32 ms nominal)
	00100  = 1:512 (Interval 16 ms nominal)
	00011  = 1:256 (Interval 8 ms nominal)
	00010  = 1:128 (Interval 4 ms nominal)
	00001  = 1:64 (Interval 2 ms nominal)
	00000  = 1:32 (Interval 1 ms nominal)
	*/
	WDTCONbits.WDTPS = uWdtPs;
	SWDTEN = 1;
	CLRWDT();
	SLEEP();
	SWDTEN = 0;
}

//-----------------------------------------------------------------

// ●割込み
void __interrupt() Interrupt(void)
{
	if (IOCIF)
	{
		IOCCFbits.IOCCF3 = 0;
	}
}
